/*******************************************************************
*******  Copyright (c) 2016-2017  Alexandros V. Gerbessiotis       *
******* All rights reserved; read copyag17.txt file                *
*******************************************************************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "avgai.h"
#include "aimisc.h"

/* operators for bspfold */

void foldmax(double *z, double *x, double *y)
{
  *z=(((*x)>=(*y))?(*x):(*y));
}

void foldmin(double *z, double *x, double *y)
{
  *z=(((*x)<(*y))?(*x):(*y));
}

void foldsum(int *z,int *x,int *y)
{
  *z= *x + *y;
}

/* in prdx1.c
void oper_add(idata *result,idata *left,idata *right,int size)
{
        register int    i;

        for (i=0;i<size;i++)
                result[i]=left[i]+right[i];
}
*/


/* One superstep broadcasting to communicate command line args */
void bspbroad(int fromp, char *from, char *to, int size)
{
 int pid;
 int nprocs;
 int i;

 pid=AIPID(); /* processor id */
 nprocs=AINPROCS(); /* number of processors */
 AIREGISTER(to,size);
 AIINIT();
 AICOMMIT();
 for (i = 0; i < nprocs;i++) {
   if (pid==fromp)
        AIHPPUT(i,from,to,0,size);
 }
 AISYNC();  /* Insert? AIBARRIER() before next line? */
 AIDEREGISTER((void*)to);
}

/* One superstep scatter (special) */
void bspscatter(int fromp, char *from, char *to, int size)
{
 int pid;
 int nprocs;
 int i;

 pid=AIPID(); /* processor id */
 nprocs=AINPROCS(); /* number of processors */
 AIREGISTER(to,size);
 AIINIT();
 AICOMMIT();
 for (i = 0; i < nprocs;i++) {
   if (pid==fromp)
        AIHPPUT(i,from+i*(size),to,0,size);
 }
 AISYNC(); /* Insert? AIBARRIER() before next line? */
 AIDEREGISTER((void*)to);
}

/* One superstep gather (special) */
void bspgather(int top, char *from, char *to, int size)
{
 int pid;
 int nprocs;
 int i;

 pid=AIPID(); /* processor id */
 nprocs=AINPROCS(); /* number of processors */
 AIREGISTER(to,nprocs*size);
 AIINIT();
 AICOMMIT();
 for (i = 0; i < nprocs;i++) {
   if (pid==i)
        AIHPPUT(0,from,to,i*size,size);
 }
 AISYNC(); /* Insert? AIBARRIER() before next line? */
 AIDEREGISTER((void*)to);
}



/* One superstep folding operation to compute global min/max of timing data */
void 
  bspfold(void (*oper)(void *,void*,void*),char *from,char *to,int size)
{
 int  pid; 
 int  nprocs; 
 int  i;
 char *list;
 int  lsize;

 list = NULL;
 pid=AIPID(); 
 nprocs=AINPROCS(); 
 if (1==nprocs)  /* if local, copy */ {
    memcpy(to,from,size*sizeof(char));
    return;
 }
 else {    /* if global, allocate memory O(p) per processor */
    lsize = nprocs*size; 
    list=(char *) malloc(lsize*sizeof(char));
    AIREGISTER(list,lsize);
    AIINIT();
    AICOMMIT();
 }
 /* each proc sends its from to every other proc; results stored in list */
 for (i = 0; i < nprocs;i++) {  
   AIHPPUT(i,from,list,pid*size,size);
 }
 AISYNC(); /* Insert? AIBARRIER() before next line? */
 /* apply operator to local list */
 for (i = 0; i < nprocs-1;i++) {
  oper(to,(list+i*size),(list+(i+1)*size));
  memcpy((list+(i+1)*size),to,size);
 }
    /* free memory and deregister */
 
    AIDEREGISTER(list);
    free((void *)list);
 }



void ai2scan(void (*operator)(),int multi,char *from,char *to, char *all,int nbytes)
{
 register int   pid;            /* processor id */
 register int   nprocs;         /* number of processors available */
 register int   i,size,sze,offset;    /* index variables */
 register int   ceil_temp,left_temp,floor_temp;
 char           *temp;                  /* buffer array */

 pid = AIPID(); nprocs = AINPROCS();

 /* split array of size multi into blocks of size either*/
 /* floor(multi/nprocs) or ceil(multi/nprocs). Assign:  */
 left_temp = LEFT(multi,nprocs);
 ceil_temp = CEIL(multi,nprocs);  /* to procs less than left_temp */
 floor_temp = FLOOR(multi,nprocs);/* to procs at least left_temp */

 /* allocate temp space and check */
 temp =  (char *) malloc(ceil_temp*nprocs*nbytes);
#ifdef SANITY_CHECK
 check_if_null((void *)temp,"ai2scan","temp");
#endif


 AI1REGISTER(temp,ceil_temp*nprocs*nbytes);
 AI1INIT(temp);
 AI1COMMIT();

 /* In first phase, each processor sends its i-th block to
 *  processor i  */
 for (i=0;i<nprocs;i++) {
    if (i < left_temp) {
          sze= ceil_temp; /* send a ceil_temp block */
          offset=i*sze;   /* starting at... */
    } else {
          sze= floor_temp; /* send a floor_temp block */
          offset=left_temp*ceil_temp+(i-left_temp)*sze;
    }
    AIOHPPUT(i,&from[offset*nbytes],temp,pid*sze*nbytes,sze*nbytes);
 }
 AI1SYNC(temp); 
 AI1BARRIER();
 /* In second phase a local prefix operation is performed */

 /* Did pid get a ceil_temp or floor_temp block? */
 if (pid < left_temp)
       size=ceil_temp;
    else
       size=floor_temp;


 for(i=1;i<nprocs;i++) {
    operator(&temp[i*size*nbytes],&temp[(i-1)*size*nbytes],
                                  &temp[i*size*nbytes],size);
 }

 /* Registration for put  operations */
 /* Register if necessary. PUSHED is defined in my_bsp.h */

 /* In third phase the computed prefices are sent back  */
 /* Where does pid send this block? */
 if (pid < left_temp) {
         sze= ceil_temp;
         offset=pid*sze;
 } else {
         sze= floor_temp;
         offset=left_temp*ceil_temp+(pid-left_temp)*sze;
 }

 for (i=0;i<nprocs;i++) {
         AIOPUT(i,&temp[i*size*nbytes],to,offset*nbytes,size*nbytes);
 }
 AI1SYNC(to);
 AI1BARRIER();

 for (i=0;i<nprocs;i++) {
         AIOPUT(i,&temp[(nprocs-1)*size*nbytes],all,offset*nbytes,size*nbytes);
 }
 AI1SYNC(all);
 AI1BARRIER();

 /* free memory and deregister information */
 AI1DEREGISTER(temp);
 /* Unregister, if necessary */
 free((void*)temp);
}


/* One superstep simple scan operation */
void
  aiscan(int  *from,int  *to,int size)
{
 int  pid;
 int  nprocs;
 register int  i,k;
 int  *list;
 int  lsize;

 pid=AIPID();
 nprocs=AINPROCS();
 if (1==nprocs){  /* if local, copy */
    memcpy(to,from,size*sizeof(int));
    return;
 }
 else {    /* if global, allocate memory O(p) per processor */
    lsize = nprocs*size*sizeof(int);
    list=(int  *) malloc(lsize);
    AIREGISTER(list,lsize);
    AIINIT();
    AICOMMIT();
 }
 /* each proc sends its from to every other proc; results stored in list */
 for (i = 0; i < nprocs;i++) {
   AIPUT(i,(char *)from,(char *)list,pid*size*sizeof(int),size*sizeof(int));
 }
 AISYNC();
 AIBARRIER();
 /* apply operator to local list */
  for(k=0;k<nprocs;k++) {
    to[k]=list[k];
  }
   for (i = 1; i <= pid;i++) {
     for(k=0;k<nprocs;k++) {
       to[k] =to[k]+list[i*size+k];
     }
   }
    /* free memory and deregister */

    AIDEREGISTER(list);
    free((void *)list);
}
